# Copyright (C) 2017-present Pinterest Inc.
#
# This module is part of ptracer and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


import ctypes
import sys


PLATFORM = None
WORD_SIZE = ctypes.sizeof(ctypes.c_void_p)
BITS = WORD_SIZE * 8

if sys.platform.startswith('linux'):
    PLATFORM = 'linux'

if PLATFORM is None or BITS != 64:
    raise RuntimeError('unsupported platform: {} ({} bit)'.format(
        sys.platform, BITS))



# Copyright (C) 2017-present Pinterest Inc.
#
# This module is part of ptracer and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0

from .ptrace import *  # noqa
from .syscalls import syscall_enter, syscall_exit  # noqa



# Copyright (C) 2017-present Pinterest Inc.
#
# This module is part of ptracer and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


from ._gen_defs_linux_64 import *  # noqa


FUNCTION_REG = 'orig_rax'
ARGS_REGS = ('rdi', 'rsi', 'rdx', 'r10', 'r8', 'r9')
RETURN_REG = 'rax'



# Copyright (C) 2017-present Pinterest Inc.
#
# This module is part of ptracer and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


import functools
import operator


class SysCallPattern(object):
    def __init__(self, name=None, args=None, result=None):
        self.name = name
        self.args = args
        self.result = result

        self.matcher = []

        if name is not None:
            self.matcher.append(self._get_comparator(
                operator.attrgetter('_name'), name))

        if result is not None:
            self.matcher.append(self._get_comparator(
                operator.attrgetter('result'), result))

        if args:
            def arg_getter(call, argno):
                return call.args[argno]

            for i, arg in enumerate(args):
                if arg is None:
                    continue

                indirection = functools.partial(arg_getter, argno=i)
                self.matcher.append(self._get_comparator(indirection, arg))

    def _get_comparator(self, indirection, value):
        if callable(value):
            checker = value
            getter = indirection
        elif hasattr(value, 'match'):
            checker = value.match
            getter = lambda sc: '{}'.format(indirection(sc).value)
        else:
            checker = lambda v: v == value
            getter = lambda sc: indirection(sc).value

        return getter, checker

    def match(self, syscall):
        return all(m[1](m[0](syscall)) for m in self.matcher)



# Copyright (C) 2017-present Pinterest Inc.
#
# This module is part of ptracer and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


import ctypes

from . import defs
from . import memory
from . import syscalldef


_ulong_t = syscalldef.CType(
    names=['unsigned', 'long'], ctype=ctypes.c_ulong, ptr_indirection=0)

_unknown_syscall = syscalldef.SysCallSig(
    name='unknown',
    params=[
        syscalldef.SysCallParamSig(
            name='param{}'.format(i),
            type=_ulong_t
        ) for i in range(len(defs.ARGS_REGS))
    ],
    result=_ulong_t
)


def syscall_enter(pid, regs, mem_fd=None):
    syscall_num = getattr(regs, defs.FUNCTION_REG)
    syscall_name = defs.SYSCALL_NUMBERS.get(
        syscall_num, '<{}>'.format(syscall_num))

    signature = defs.SYSCALLS.get(syscall_name, _unknown_syscall)

    args = []

    for i, param in enumerate(signature.params):
        raw_value = getattr(regs, defs.ARGS_REGS[i])
        ptype = param.type

        if ptype.ptr_indirection or issubclass(ptype.ctype, ctypes.Array):
            if raw_value != 0:
                value = memory.read_c_type_ptr(
                    pid, raw_value, ptype.ctype, ptype.ptr_indirection, mem_fd)
            else:
                value = None
        else:
            value = ptype.ctype(raw_value).value

        arg = syscalldef.SysCallArg(name=param.name, type=param.type,
                                    raw_value=raw_value, value=value)

        args.append(arg)

    syscall = syscalldef.SysCall(
        name=syscall_name, args=args, result=None, pid=pid)

    return syscall


def syscall_exit(syscall, regs, mem_fd=None):
    signature = defs.SYSCALLS.get(syscall.name, _unknown_syscall)

    for i, param in enumerate(signature.params):
        ptype = param.type
        if (not ptype.ptr_indirection and
                not issubclass(ptype.ctype, ctypes.Array)):
            continue

        raw_value = getattr(regs, defs.ARGS_REGS[i])
        if raw_value != 0:
            value = memory.read_c_type_ptr(
                syscall.pid, raw_value, ptype.ctype,
                ptype.ptr_indirection, mem_fd)
        else:
            value = None

        arg = syscall.args[i]
        arg.raw_value = raw_value
        arg.value = value

    restype = signature.result
    raw_result = getattr(regs, defs.RETURN_REG)

    if restype.ptr_indirection:
        if raw_result != 0:
            value = memory.read_c_type_ptr(syscall.pid, raw_result, restype,
                                           restype.ptr_indirection)
        else:
            value = None
    else:
        value = signature.result.ctype(raw_result).value

    syscall.result = syscalldef.SysCallResult(restype, raw_result, value)

    return syscall



# Copyright (C) 2017-present Pinterest Inc.
#
# This module is part of ptracer and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


from __future__ import print_function

import errno
import fcntl
import linecache
import logging
import multiprocessing
import os
try:
    import Queue as queue
except ImportError:
    import queue
import select
import signal
import struct
import threading
import time
import traceback

from ptracer import _lltraceback
from ptracer import ptrace


logger = logging.getLogger('ptracer')


class PtracerError(Exception):
    def __init__(self, msg, orig_exc=None):
        super(PtracerError, self).__init__(msg)
        self.orig_exc = orig_exc

    def __str__(self):
        if self.orig_exc is not None:
            return '{}\n{}'.format(self.args[0], self.orig_exc)
        else:
            return self.args[0]


# The tracing thread is run parallel to the traced thread,
# the syscall callback will be called in this thread.
def _tracing_thread(handler_cb, thread_stop_event, debugger_start_event,
                    thread_map, syscall_filter, error_queue):

    # The main syscall queue.
    syscall_queue = multiprocessing.Queue()

    # Pipes for communication with lltraceback helper.
    stack_response_read, stack_response_write = os.pipe()
    stack_request_read, stack_request_write = os.pipe()

    # Debugger process start request event.
    dbgproc_start = multiprocessing.Event()
    # Debugger process stop request event.
    dbgproc_stop = multiprocessing.Event()
    # Debugger process started response event.
    dbgproc_started = multiprocessing.Event()

    # The actual tracing is done by a subprocess.
    # It is necessary because of the GIL, as a ptrace-stopped thread
    # holding the GIL would block the tracer thread as well.
    ptrace_process = multiprocessing.Process(
        target=_tracing_process,
        args=(os.getpid(), dbgproc_started, dbgproc_start, dbgproc_stop,
              stack_request_write, stack_response_read,
              syscall_queue, syscall_filter, error_queue))

    ptrace_process.start()

    if hasattr(ptrace, 'set_ptracer'):
        # On systems with Yama LSM enabled in mode 1 (e.g. Ubuntu Trusty),
        # PTRACE_ATTACH will fail with EPERM unless we call PR_SET_PTRACER
        # with the PID of the tracing process.
        try:
            ptrace.set_ptracer(ptrace_process.pid)
        except OSError as e:
            pass

    dbgproc_start.set()

    # The lltraceback thread is a low-level GIL-independent thread
    # that is used to dump the current call stack in a given Python thread.
    _lltraceback.start_thread(
        stack_request_read, stack_response_write, thread_map)

    try:
        # Wait for debugger to start
        if not dbgproc_started.wait(1):
            # If the debugger has not started in 1 second, assume
            # it died and bail out.
            return

        # Notify the main thread that we're ready.
        debugger_start_event.set()

        while True:
            if thread_stop_event.is_set() or not ptrace_process.is_alive():
                # The tracing context has exited, stop the debugger.
                dbgproc_stop.set()

                while True:
                    # Drain the syscall queue.
                    try:
                        syscall = syscall_queue.get(timeout=0.1)
                    except queue.Empty:
                        break
                    else:
                        handler_cb(syscall)
                break

            try:
                event = syscall_queue.get_nowait()
            except queue.Empty:
                time.sleep(0.05)
                continue

            try:
                handler_cb(event)
            except Exception as e:
                logger.exception('EXCEPTION IN SYSCALL CALLBACK')
    finally:
        dbgproc_stop.set()
        _lltraceback.stop_thread()
        ptrace_process.join(1)
        if ptrace_process.exitcode is None:
            ptrace_process.terminate()
            ptrace_process.join(1)
            if ptrace_process.exitcode is None:
                os.kill(ptrace_process.pid, signal.SIGKILL)

        os.close(stack_response_read)
        os.close(stack_response_write)
        os.close(stack_request_read)
        os.close(stack_request_write)


def _tracing_process(pid, dbgproc_started, dbgproc_start, dbgproc_stop,
                     stack_request_pipe, stack_response_pipe,
                     syscall_queue, syscall_filter, error_queue):
    # The tracing process consists of two threads:
    # the first reads the call stacks from the tracee, and the second
    # does the actual tracing.

    if not dbgproc_start.wait(1):
        # The parent failed to continue the startup.
        err = PtracerError('Debugger startup handshake failed')
        error_queue.put_nowait(err)
        return

    stack_queue = queue.Queue()

    dbgthread_stop = threading.Event()

    debugger_thread = threading.Thread(
        target=_debugger_thread,
        args=(pid, dbgproc_started, dbgthread_stop,
              stack_request_pipe, stack_queue,
              syscall_queue, syscall_filter, error_queue),
        name='pytracer-debugger')

    debugger_thread.daemon = True
    debugger_thread.start()

    try:
        _read_callstacks(stack_response_pipe, stack_queue, debugger_thread,
                         dbgproc_stop)

    except Exception:
        logger.debug('Unhandled exception in ptrace process', exc_info=True)
        err = PtracerError('Unhandled exception in ptrace process',
                           orig_exc=traceback.format_exc())
        error_queue.put_nowait(err)

    finally:
        syscall_queue.close()

        if debugger_thread.is_alive():
            # Unblock the debugger if it is waiting on the stack queue
            stack_queue.put_nowait(None)
            dbgthread_stop.set()
            debugger_thread.join()


def _read_callstacks(stack_response_pipe, stack_queue, debugger_thread,
                     dbgproc_stop):
    buf = b''
    stacklen = -1
    tuplesize = 0
    elemlen = -1
    required_len = 4
    stack = []
    entry = []

    fcntl.fcntl(stack_response_pipe, fcntl.F_SETFL, os.O_NONBLOCK)

    # The call stack format is as follows:
    #     stack_length:uint32_t
    #     entry_tuple_len:uint32_t
    #     (
    #            (item_length:uint32_t
    #             item_data:char[item_length]) * entry_tuple_len
    #     ) * stack_length
    while True:
        ready, _, _ = select.select([stack_response_pipe], [], [], 1.0)

        if ready:
            try:
                buf += os.read(stack_response_pipe, 4096)
            except OSError as e:
                if e.errno == errno.EAGAIN:
                    pass
                else:
                    raise
            else:
                while len(buf) >= required_len:
                    if stacklen == -1:
                        stacklen = struct.unpack('!i', buf[:4])[0]
                        buf = buf[4:]
                        if stacklen == 0:
                            # No stack could be extracted.
                            stack_queue.put_nowait([])
                            stacklen = -1

                    elif tuplesize == 0:
                        tuplesize = struct.unpack('!i', buf[:4])[0]
                        buf = buf[4:]

                    elif elemlen == -1:
                        elemlen = struct.unpack('!i', buf[:4])[0]
                        buf = buf[4:]
                        required_len = elemlen

                    else:
                        elem = buf[:elemlen]
                        buf = buf[elemlen:]

                        if len(entry) == 1:
                            elem = int(elem)
                        else:
                            elem = elem.decode('utf-8')
                        entry.append(elem)
                        if len(entry) == tuplesize:
                            if tuplesize == 3:
                                entry.append(
                                    linecache.getline(entry[0], entry[1]))

                            stack.append(tuple(entry))
                            entry = []
                            if len(stack) == stacklen:
                                stack_queue.put_nowait(list(reversed(stack)))
                                stack = []
                                stacklen = -1
                                tuplesize = 0

                        elemlen = -1
                        required_len = 4

        if not debugger_thread.is_alive():
            # The debugger thread has stopped.
            break

        if dbgproc_stop.is_set():
            # We were asked to stop by the traced process.
            break


def _debugger_thread(main_pid, dbgproc_started, dbgthread_stop,
                     stack_request_pipe, stack_queue,
                     syscall_queue, syscall_filter, error_queue):
    try:
        _debugger_thread_inner(main_pid, dbgproc_started, dbgthread_stop,
                               stack_request_pipe, stack_queue, syscall_queue,
                               syscall_filter)
    except Exception:
        logger.debug('Unhandled exception in ptrace process', exc_info=True)
        err = PtracerError('Unhandled exception in ptrace process',
                           orig_exc=traceback.format_exc())
        error_queue.put_nowait(err)


def _debugger_thread_inner(main_pid, dbgproc_started, dbgthread_stop,
                           stack_request_pipe, stack_queue,
                           syscall_queue, syscall_filter):
    ptrace_options = ptrace.PTRACE_O_TRACECLONE
    # Attach to the tracee and wait for it to stop.
    ptrace.attach_and_wait(main_pid, ptrace_options)

    if syscall_filter is not None:
        filter_ = lambda sc: any(m.match(sc) for m in syscall_filter)
    else:
        filter_ = None

    syscall_trap = signal.SIGTRAP | 0x80
    enabled = False
    signum = 0
    syscall_state = {}
    sigstop_received = set()

    processes = {main_pid}
    mem_fds = {}
    mem_fds[main_pid] = _open_procmem(main_pid)

    # Notify the parent that we are ready to start tracing.
    dbgproc_started.set()

    try:
        # Restart the tracee and enter the tracing loop.
        ptrace.syscall(main_pid)

        while True:
            if dbgthread_stop.is_set():
                break

            pid, status = ptrace.wait(-1)

            if os.WIFEXITED(status) or os.WIFSIGNALED(status):
                # Traced thread has died.
                processes.discard(pid)
                mem_fd = mem_fds.get(pid)
                if mem_fd is not None:
                    try:
                        os.close(mem_fd)
                    except IOError:
                        pass
                if not processes:
                    break
                else:
                    continue

            elif os.WIFSTOPPED(status):
                ptrace_event = ptrace.WPTRACEEVENT(status)
                if ptrace_event == ptrace.PTRACE_EVENT_CLONE:
                    # A new thread has been created.
                    new_pid = ptrace.geteventmsg(pid)
                    # See the comment below for the explanation of this check.
                    if new_pid not in sigstop_received:
                        ptrace.wait_for_trace_stop(new_pid)
                        try:
                            ptrace.syscall(new_pid)
                        except OSError as e:
                            if e.errno != errno.ESRCH:
                                # The new thread might have already died.
                                raise
                    else:
                        sigstop_received.discard(new_pid)

                    mem_fds[new_pid] = _open_procmem(new_pid)

                    processes.add(new_pid)
                    ptrace.syscall(pid)
                    continue

                stopsig = os.WSTOPSIG(status)
                if stopsig != syscall_trap:
                    # Signal-delivery-stop.

                    # The special condition below is for cases when we
                    # receive a SIGSTOP for a newly created thread _before_
                    # receiving the PTRACE_EVENT_CLONE event for its parent.
                    # In this case we must not forward the signal, but
                    # must record its receipt so that once we _do_ receive
                    # PTRACE_EVENT_CLONE for the parent, we don't wait for
                    # SIGSTOP in the child again.
                    if (stopsig != signal.SIGSTOP or
                            pid in processes or
                            all(syscall.name != 'clone'
                                for syscall in syscall_state.values()
                                if syscall is not None)):
                        # forward the signal
                        signum = stopsig
                    else:
                        sigstop_received.add(pid)
                else:
                    # Syscall-stop.
                    syscall = syscall_state.get(pid)
                    regs = ptrace.getregs(pid)
                    mem_fd = mem_fds.get(pid)

                    if syscall is None:
                        # Syscall-enter-stop.
                        syscall_state[pid] = ptrace.syscall_enter(
                            pid, regs, mem_fd)
                    else:
                        # Syscall-exit-stop.
                        ptrace.syscall_exit(syscall, regs, mem_fd)

                        if enabled:
                            # Stop tracing once the tracee executes
                            # the magic open() in ptracer.disable().
                            stop_tracing = (
                                syscall.name == 'open' and
                                syscall.args[0].value == b'\x03\x02\x01'
                            ) or (
                                syscall.name == 'openat' and
                                syscall.args[1].value == b'\x03\x02\x01'
                            )

                            if stop_tracing:
                                break
                            elif filter_ is None or filter_(syscall):
                                # Wait for the traceback to arrive.
                                os.write(stack_request_pipe,
                                         struct.pack('!Q', pid))
                                stack = stack_queue.get()
                                if stack is None:
                                    ptrace.cont(pid)
                                    break

                                syscall.traceback = stack
                                syscall_queue.put_nowait(syscall)

                        elif not enabled:
                            # Start tracing once the tracee executes
                            # the magic open() in ptracer.enable().
                            start_tracing = (
                                syscall.name == 'open' and
                                syscall.args[0].value == b'\x01\x02\x03'
                            ) or (
                                syscall.name == 'openat' and
                                syscall.args[1].value == b'\x01\x02\x03'
                            )

                            if start_tracing:
                                enabled = True

                        syscall_state[pid] = None
            else:
                logger.error('unexpected status of traced process %s: %s',
                             pid, status)

            # Continue until next syscall.
            ptrace.syscall(pid, signum)
            signum = 0
    finally:
        for process in processes:
            try:
                ptrace.detach(process)
            except OSError as e:
                if e.errno == errno.ESRCH:
                    pass
                else:
                    raise

        for fd in mem_fds.values():
            try:
                os.close(fd)
            except (OSError, IOError):
                pass


def _open_procmem(pid):
    try:
        mem_fd = os.open('/proc/{}/mem'.format(pid), os.O_RDONLY)
    except IOError as e:
        if e.errno == errno.EACCESS:
            logger.debug('cannot access /proc/{}/mem'.format(pid),
                         exc_info=True)
            return None
        else:
            raise
    else:
        return mem_fd



# Copyright (C) 2017-present Pinterest Inc.
#
# This module is part of ptracer and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


class CType(object):
    __slots__ = ('names', 'ctype', 'ptr_indirection')

    def __init__(self, names, ctype, ptr_indirection):
        self.names = names
        self.ctype = ctype
        self.ptr_indirection = ptr_indirection

    def __repr__(self):
        return '<CType {}>'.format(' '.join(self.names))


class SysCallParamSig(object):
    __slots__ = ('name', 'type')

    def __init__(self, name, type):
        self.name = name
        self.type = type

    def __repr__(self):
        return '<SysCallParamSig {} {}>'.format(self.type, self.name)


class SysCallSig(object):
    __slots__ = ('name', 'params', 'result')

    def __init__(self, name, params, result):
        self.name = name
        self.params = params
        self.result = result


class SysCallArg(object):
    def __init__(self, name, type, raw_value, value):
        self.name = name
        self.type = type
        self.raw_value = raw_value
        self.value = value

    def __repr__(self):
        return '<SysCallArg {}={!r}>'.format(self.name, self.value)


class SysCallResult(object):
    def __init__(self, type, raw_value, value):
        self.type = type
        self.raw_value = raw_value
        self.value = value

    def __repr__(self):
        return '<SysCallResult {!r}>'.format(self.value)


class SysCall(object):
    def __init__(self, name, args, result, pid, traceback=None):
        self.name = name
        self._name = SysCallArg(None, None, self.name, self.name)
        self.pid = pid
        self.args = args
        self.result = result
        self.traceback = traceback

    def __repr__(self):
        return '<SysCall {!s}>'.format(self.name)



# Copyright (C) 2017-present Pinterest Inc.
#
# This module is part of ptracer and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


from . import platform
from .platform import *  # noqa


if platform.PLATFORM == 'linux':
    from ._defs_linux import *  # noqa



# Copyright (C) 2017-present Pinterest Inc.
#
# This module is part of ptracer and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


import ctypes
import errno
import os
import signal

from .defs import *  # noqa

from . import defs
from ._ptrace import ptrace as _ptrace
try:
    from ._ptrace import set_ptracer  # noqa
except ImportError:
    pass


def traceme():
    return _ptrace(defs.PTRACE_TRACEME, 0, 0, 0)


def peektext(pid, addr):
    return _ptrace(defs.PTRACE_PEEKTEXT, pid, addr, 0)


def peekdata(pid, addr):
    return _ptrace(defs.PTRACE_PEEKDATA, pid, addr, 0)


def peekuser(pid, addr):
    return _ptrace(defs.PTRACE_PEEKUSER, pid, addr, 0)


def poketext(pid, addr, data):
    return _ptrace(defs.PTRACE_POKETEXT, pid, addr, data)


def pokedata(pid, addr, data):
    return _ptrace(defs.PTRACE_POKEDATA, pid, addr, data)


def pokeuser(pid, addr, data):
    return _ptrace(defs.PTRACE_POKEUSER, pid, addr, data)


def getregs(pid):
    regs = defs.user_regs_struct()
    _ptrace(defs.PTRACE_GETREGS, pid, 0, ctypes.addressof(regs))
    return regs


def getfpregs(pid):
    regs = defs.user_fpregs_struct()
    _ptrace(defs.PTRACE_GETREGS, pid, 0, ctypes.addressof(regs))
    return regs


def getsiginfo(pid):
    siginfo = defs.siginfo_t()
    _ptrace(defs.PTRACE_GETSIGINFO, pid, 0, ctypes.addressof(siginfo))
    return siginfo


def setoptions(pid, options):
    return _ptrace(defs.PTRACE_SETOPTIONS, pid, 0, options)


def geteventmsg(pid):
    data = ctypes.c_ulong()
    _ptrace(defs.PTRACE_GETEVENTMSG, pid, 0, ctypes.addressof(data))
    return data.value


def cont(pid, signum=0):
    return _ptrace(defs.PTRACE_CONT, pid, 0, signum)


def syscall(pid, signum=0):
    return _ptrace(defs.PTRACE_SYSCALL, pid, 0, signum)


def kill(pid):
    return _ptrace(defs.PTRACE_KILL, pid, 0, 0)


def attach(pid):
    return _ptrace(defs.PTRACE_ATTACH, pid, 0, 0)


def attach_and_wait(pid, options=0):
    attach(pid)
    wait_for_trace_stop(pid)
    options |= defs.PTRACE_O_TRACESYSGOOD
    setoptions(pid, options)


def wait_for_trace_stop(pid):
    try:
        _wait_for_trace_stop(pid)
    except BaseException:
        try:
            # If _wait_for_trace_stop fails for any reason,
            # we must try to detach from the tracee to avoid
            # leaving it blocked.
            detach(pid)
        except BaseException:
            pass

        raise


def _wait_for_trace_stop(pid):
    try:
        # First, check if the tracee is already stopped.
        siginfo = getsiginfo(pid)
    except OSError as e:
        if e.errno == errno.ESRCH:
            # The tracee is still running, so we'll wait
            pass
        else:
            raise
    else:
        # Normally, PTRACE_ATTACH will send a SIGSTOP to the tracee,
        # which we will see here.  However, on some kernels the actual
        # signal may sometimes be SIGTRAP, and that seems to happen
        # when the previous tracer had died without calling PTRACE_DETACH
        # on this process first.  In this case, we need to restart the process
        # and wait for the real SIGSTOP.
        if siginfo.si_signo == signal.SIGTRAP:
            cont(pid, siginfo.si_signo)
        elif is_stop_signal(siginfo.si_signo):
            return
        else:
            raise OSError('traced process has stopped with an unexpected '
                          'signal {}'.format(siginfo.si_signo))

    pid, status = wait(pid)

    if os.WIFEXITED(status):
        raise OSError('traced process {} has exited with exit code {}'.format(
            pid, os.WEXITSTATUS(status)))

    elif os.WIFSIGNALED(status):
        raise OSError('traced process {} has been killed by '
                      'the {} signal {}'.format(pid, os.WTERMSIG(status)))

    if not os.WIFSTOPPED(status):
        raise OSError('waitpid({}) returned an unexpected status {}'.format(
            pid, hex(status)))

    stopsig = os.WSTOPSIG(status)
    if stopsig != signal.SIGSTOP:
        raise OSError('waitpid({}) returned an unexpected status {}'.format(
            pid, hex(status)))


def wait(pid, options=0):
    options |= defs.WALL
    return os.waitpid(pid, options)


def detach(pid, signum=0):
    return _ptrace(defs.PTRACE_DETACH, pid, 0, signum)


def is_stop_signal(signum):
    return signum in (signal.SIGSTOP, signal.SIGTSTP,
                      signal.SIGTTIN, signal.SIGTTOU)



# Copyright (C) 2017-present Pinterest Inc.
#
# This module is part of ptracer and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


__version__ = '0.6'
__all__ = ['context', 'enable', 'disable', 'SysCallPattern']


try:
    import Queue as queue
except ImportError:
    import queue
import multiprocessing
import threading

from . import _ptracer
from ._ptracer import PtracerError  # NOQA
from ._syscall import SysCallPattern  # NOQA
from . import _lltraceback


class TracingContext(object):
    def __init__(self):
        self.enabled = False

    def enable(self, handler_cb, filter=None):
        if self.enabled:
            raise RuntimeError('tracing context is already enabled')

        self.enabled = True
        self.thread_stop_event = threading.Event()

        debugger_start_event = threading.Event()

        # Debugger error queue.
        self.error_queue = multiprocessing.Queue()

        if isinstance(filter, SysCallPattern):
            filter = [filter]

        self.ptrace_thread = threading.Thread(
            target=_ptracer._tracing_thread,
            args=(handler_cb, self.thread_stop_event, debugger_start_event,
                  {_lltraceback.gettid(): threading.current_thread().ident},
                  filter, self.error_queue))

        self.ptrace_thread.start()

        # Wait for debugger to start
        if not debugger_start_event.wait(1):
            try:
                self.disable()
            except Exception:
                raise
            else:
                raise PtracerError('Unhandled exception in ptrace process')
        else:
            try:
                # Perform a magic syscall to enable
                # syscall callback invocation.
                open(b'\x01\x02\x03', 'r')
            except IOError:
                pass

    def disable(self):
        if not self.enabled:
            return

        try:
            # Notify the debugger we're not tracing anymore.
            open(b'\x03\x02\x01', 'r')
        except IOError:
            pass

        self.enabled = False
        self.thread_stop_event.set()
        self.thread_stop_event = None
        self.ptrace_thread.join()
        self.ptrace_thread = None

        try:
            error = self.error_queue.get_nowait()
        except queue.Empty:
            error = None

        self.error_queue.close()
        self.error_queue = None

        if error is not None:
            raise error


class context(object):
    """Tracing context manager."""

    def __init__(self, handler_cb, filter=None):
        self.handler_cb = handler_cb
        self.filter = filter

    def __enter__(self):
        _context.enable(self.handler_cb, filter=self.filter)

    def __exit__(self, exc_type, exc_value, exc_tb):
        _context.disable()


_context = TracingContext()
enable = _context.enable
disable = _context.disable



# Copyright (C) 2017-present Pinterest Inc.
#
# This module is part of ptracer and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


import ctypes
import os
import signal

from . import platform
from . import syscalldef

if platform.BITS == 64:
    from ._defs_linux_64 import *  # noqa
else:
    raise RuntimeError('unsupported platform: {} ({} bit)'.format(
        platform.PLATFORM, platform.BITS))

from ._gen_defs_linux_64 import SYSCALLS


WALL = 0x40000000


# Must not intepret the fifth arg of futext() as a pointer
# because it might be a value:
#     int futex(int *uaddr, int futex_op, int val,
#               const struct timespec *timeout,   /* or: uint32_t val2 */
#               int *uaddr2, int val3);
SYSCALLS['futex'].params[3] = syscalldef.SysCallParamSig(
    name='val2', type=syscalldef.CType(['uint32_t'], ctypes.c_uint32, 0))
SYSCALLS['futex'].params[4] = syscalldef.SysCallParamSig(
    name='uaddr2', type=syscalldef.CType(['void', '*'], ctypes.c_void_p, 0))


class c_int_Array_2(ctypes.Array):
    _length_ = 2
    _type_ = ctypes.c_int


SYSCALLS['pipe'].params[0] = syscalldef.SysCallParamSig(
    name='pipefd', type=syscalldef.CType(['int', '[2]'], c_int_Array_2, 0))
SYSCALLS['pipe2'].params[0] = syscalldef.SysCallParamSig(
    name='pipefd', type=syscalldef.CType(['int', '[2]'], c_int_Array_2, 0))


def WPTRACEEVENT(status):
    if os.WIFSTOPPED(status):
        stopsig = os.WSTOPSIG(status)
        if stopsig == signal.SIGTRAP:
            return status >> 16

    return 0



import os
import os.path
import sys

import setuptools


CFLAGS = ['-Wall', '-Wsign-compare', '-Wconversion']


if sys.platform in ('win32', 'cygwin', 'cli'):
    raise RuntimeError('ptracer is a Unix-only library')

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
    readme = f.read()

with open(os.path.join(
        os.path.dirname(__file__), 'ptracer', '__init__.py')) as f:
    for line in f:
        if line.startswith('__version__ ='):
            _, _, version = line.partition('=')
            VERSION = version.strip(" \n'\"")
            break
    else:
        raise RuntimeError(
            'unable to read the version from ptracer/__init__.py')

setuptools.setup(
    name='ptracer',
    version=VERSION,
    description='On-demand system call tracing for Python programs.',
    long_description=readme,
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Operating System :: POSIX :: Linux',
        'Development Status :: 4 - Beta',
    ],
    platforms=['POSIX'],
    license='Apache License, Version 2.0',
    provides=['ptracer'],
    packages=['ptracer', 'ptracer.ptrace'],
    ext_modules=[
        setuptools.Extension(
            'ptracer._lltraceback',
            ['ptracer/_lltraceback.c'],
            extra_compile_args=CFLAGS,
        ),
        setuptools.Extension(
            'ptracer.ptrace._ptrace',
            ['ptracer/ptrace/_ptrace.c'],
            extra_compile_args=CFLAGS,
        )
    ],
    test_suite='tests.suite',
)



# Copyright (C) 2017-present Pinterest Inc.
#
# This module is part of ptracer and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


import ctypes
import os
import struct

from . import defs
from . import ptrace


_CONTAINER_TYPES = (ctypes.Array, ctypes.Structure, ctypes.Union)


def _ptrace_read_word(pid, address):
    word = ptrace.peektext(pid, address)
    return struct.pack('q', word)


def ptrace_read(pid, address, bytecount):
    # PTRACE_PEEK reads must be aligned on the word boundary.
    lpad = address % defs.WORD_SIZE
    rpad = (bytecount + lpad) % defs.WORD_SIZE
    if rpad:
        rpad = defs.WORD_SIZE - rpad

    data = bytearray(bytecount)
    offset = 0

    if lpad:
        word = _ptrace_read_word(pid, address - lpad)
        chunk = word[lpad:]
        chunk_len = len(chunk)
        data[offset:offset + chunk_len] = word[lpad:]
        offset += chunk_len
        bytecount -= chunk_len

    wordcount = (bytecount + rpad) // defs.WORD_SIZE
    for _ in range(wordcount - 1):
        word = _ptrace_read_word(pid, address + offset)
        data[offset:offset + defs.WORD_SIZE] = word
        offset += defs.WORD_SIZE

    if wordcount:
        word = _ptrace_read_word(pid, address - lpad)
        if rpad:
            data[offset:offset + defs.WORD_SIZE - rpad] = word[:-rpad]
        else:
            data[offset:offset + defs.WORD_SIZE] = word

    return data


def procmem_read(fd, address, bytecount):
    os.lseek(fd, address, os.SEEK_SET)
    return bytearray(os.read(fd, bytecount))


def read_c_type_ptr(pid, address, c_type, indirection=1, mem_fd=None):
    if indirection > 1:
        address = read_c_type_ptr(pid, address, ctypes.c_void_p,
                                  indirection - 1)

    if issubclass(c_type, ctypes.c_char):
        if mem_fd is not None:
            try:
                return procmem_read_c_string(mem_fd, address)
            except IOError:
                return ptrace_read_c_string(pid, address)
        else:
            return ptrace_read_c_string(pid, address)
    else:
        bytecount = ctypes.sizeof(c_type)
        if mem_fd is not None:
            try:
                data = procmem_read(mem_fd, address, bytecount)
            except IOError:
                data = ptrace_read(pid, address, bytecount)
        else:
            data = ptrace_read(pid, address, bytecount)

        c_value = c_type.from_buffer(data)

        if not issubclass(c_type, _CONTAINER_TYPES):
            value = c_value.value
        elif issubclass(c_type, ctypes.Array):
            value = tuple(c_value)
        else:
            value = c_value

        return value


def ptrace_read_c_string(pid, address, max_size=1024):
    # PTRACE_PEEK reads must be aligned on the word boundary.
    bytecount = max_size
    lpad = address % defs.WORD_SIZE
    rpad = (bytecount + lpad) % defs.WORD_SIZE
    if rpad:
        bytecount += defs.WORD_SIZE - rpad

    data = bytearray(bytecount)
    offset = 0

    if lpad:
        word = _ptrace_read_word(pid, address - lpad)
        chunk = word[lpad:]
        nulpos = chunk.find(b'\x00')
        if nulpos != -1:
            return chunk[:nulpos]

        chunk_len = len(chunk)
        data[offset:offset + chunk_len] = word[lpad:]
        offset += chunk_len
        bytecount -= chunk_len

    wordcount = (bytecount + rpad) // defs.WORD_SIZE
    for _ in range(wordcount):
        word = _ptrace_read_word(pid, address + offset)
        nulpos = word.find(b'\x00')
        if nulpos != -1:
            data[offset:offset + nulpos] = word[:nulpos]
            return bytes(data[:offset + nulpos])
        else:
            chunk_len = defs.WORD_SIZE

        data[offset:offset + defs.WORD_SIZE] = word
        offset += defs.WORD_SIZE

    return bytes(data)


def procmem_read_c_string(fd, address, max_size=1024):
    data = procmem_read(fd, address, max_size)
    nulpos = data.find(b'\x00')
    if nulpos != -1:
        return data[:nulpos]
    else:
        return data
